#include "BF.h"

#include <stdio.h>
#include <map>

namespace BF
{
    const double BF::BF_EPS = 1e-10;
    
    BF::BF(int _numFreq, int _samplingRate, double _soundSpeed) : numFreq(_numFreq), samplingRate(_samplingRate), soundSpeed(_soundSpeed), transformFunctionDatabase(NULL)
    {
    }

    BF::~BF()
    {
	delete transformFunctionDatabase;
    }

    void
    BF::calculateTransformFunctions(const std::vector<MicLocation> &micPositions, const std::vector<SourceLocation> &sourcePositions, std::vector<Matrix> &transformFunctions) const
    {
/* == source from pos2tf.m

function H = pos2tf(Micpos, Srcpos, f, c)
% H = pos2tf(Micpos, Srcpos, f[, c])
% Micpos: Mic position [m], array (3[x; y; z], NMic)
% Srcpos: Source position [m], array (3[x; y; z], NSource)
% f: Frequency vector [Hz], vector (NFreq, 1)
% c: Sound speed [m/s], scalar (default: 340)
% H: Transfer function (TF) matrix, 3D-array (NMic, NSource, NFreq)

if nargin < 4, c = 340; end

NMic = size(Micpos, 2);
NSrc = size(Srcpos, 2);
NFreq = size(f, 1);

Dist = zeros(NMic, NSrc);
for ns = 1:NSrc
    rpos = Micpos - repmat(Srcpos(:, ns), 1, NMic);
    Dist(:, ns) = sqrt(sum(rpos.^2)); % Distance [m]
end
Del = Dist/c; % Delay time [s]

H = zeros(NMic, NSrc, NFreq);
for nf = 1:NFreq
    H(:, :, nf) = 1./Dist.*exp(-i*2*pi*f(nf)*Del);
end

*/

	int numSrc = sourcePositions.size();
	std::vector<Matrix> newTransformFunctions;
	std::vector<MicLocation> newMicPositions;
	std::vector<SourceLocation> newSourcePositions;

	if(this->transformFunctionDatabase == NULL){
	    int numMic = micPositions.size();

	    /* 伝達関数の計算 */
	    Matrix distances(numMic, numSrc);
	    for(int i = 0; i < numMic; i++){
		const Location &micPos = micPositions[i];

		//fprintf(stdout, "mic=(%f, %f, %f)\n", micPos.x, micPos.y, micPos.z);
		for(int j = 0; j < numSrc; j++){
		    const SourceLocation &srcPos = sourcePositions[j];
		    //fprintf(stdout, "src=(%f, %f, %f)\n", srcPos.x, srcPos.y, srcPos.z);
		    double d = sqrt((micPos.x - srcPos.x) * (micPos.x - srcPos.x)
				    + (micPos.y - srcPos.y) * (micPos.y - srcPos.y)
				    + (micPos.z - srcPos.z) * (micPos.z - srcPos.z));
		    distances.setElementAt(i, j, d);
		}
	    }

	    for(int i = 0; i < this->numFreq; i++){
		double f = (double)samplingRate * (i * 2 + 1) / 4 / this->numFreq;

		Matrix transformFunction(numMic, numSrc);
		for(int j = 0; j < numMic; j++){
		    for(int k = 0; k < numSrc; k++){
			double d = distances.getElementAt(j, k).real();
			double t = -2 * M_PI * f/*freq*/ * (d / soundSpeed)/*delay*/;
			double vr = cos(t) / d;
			double vi = sin(t) / d;
		    
			transformFunction.setElementAt(j, k, vr, vi);
		    }
		}

		//transformFunction.dump(stdout, "tf");
		newTransformFunctions.push_back(transformFunction);
	    }
	    
	    newMicPositions = micPositions;
	    newSourcePositions = sourcePositions;

	}else{
	    /* 伝達関数をデータベースから取得する場合、マイク数はデータベースから取得する。 */
	    int numMic = this->transformFunctionDatabase->micLocations.size();

	    /* 伝達関数をデータベースから取得するが、まずは取得結果を覚める領域を確保 */
	    for(int i = 0; i < this->numFreq; i++){
		newTransformFunctions.push_back(Matrix(numMic, numSrc));
	    }
	    /* 続いて、入力の音源・マイクの組を、データベース上どの音源・マイクの組にマップするかを検索 */
	    std::map<int, int> srcMap, micMap;
	    for(int i = 0; i < numSrc; i++){
		int srcIndex = this->transformFunctionDatabase->findNearestSourceIndex(sourcePositions[i]);
		srcMap.insert(std::pair<int, int>(i, srcIndex));
	    }

	    for(int i = 0; i < this->numFreq; i++){
		Matrix &transformFunction = newTransformFunctions[i];
		for(int j = 0; j < numSrc; j++){
		    int jj = srcMap[j];
		    transformFunction.setColumn(j, this->transformFunctionDatabase->transformFunctions[i].getColumn(srcMap[j]));
		}

		// 2012/05/24 deleted by N.Tanaka
		//newTransformFunctions.push_back(transformFunction);
	    }
	    
	}

	/* 結果の格納 */
	newTransformFunctions.swap(transformFunctions);
    }

    void
    BF::executeBeamForming(const Matrix &inputSpectrum, Matrix &outputSpectrum)
    {
	Matrix newInputSpectrum (inputSpectrum);
	Matrix newOutputSpectrum(this->separationMatrices[0].getRowCount(), this->numFreq);

	for(int i = 0; i < this->numFreq; i++){
//	    this->separationMatrices[i].dump(stdout, "this->separationMatrices[%d]", i);
//	    inputSpectrum.getColumn(i).dump(stdout, "inputSpectrum.getColumn(%d)", i);
//	    this->separationMatrices[i].multiplied(inputSpectrum.getColumn(i)).dump(stdout, "this->separationMatrices[%d].multiplied(inputSpectrum.getColumn(%d))", i, i);
	    newOutputSpectrum.setColumn(i, this->separationMatrices[i].multiplied(inputSpectrum.getColumn(i)));
	}

//	inputSpectrum.dump(stdout, "in");
//	separationMatrices[0].dump(stdout, "G[0]");
//	newOutputSpectrum.dump(stdout, "out");

	outputSpectrum = newOutputSpectrum;
	newInputSpectrum.swap(this->inputSpectrum);
	newOutputSpectrum.swap(this->outputSpectrum);
    }

    void
    BF::setTransformFunctionDatabase(const TransformFunctionDatabase &transformFunctionDatabase)
    {
	using std::swap;
	TransformFunctionDatabase *newTransformFunctionDatabase = new TransformFunctionDatabase(transformFunctionDatabase);
	swap(this->transformFunctionDatabase, newTransformFunctionDatabase);
	delete newTransformFunctionDatabase;
    }
}
